
From: Chris Mason <mason@suse.com>

find_group_other looks buggy for ext2 and ext3 in 2.6, it can cause -ENOSPC
errors when the fs has plenty of free room.

To hit the bug, you need a filesystem where:

parent_group has no free blocks (but might have free inodes) Every other
group with free inodes has no free blocks.

That gets you down to the final linear search in find_group_other.  The
linear search has two bugs:

group = parent_group + 1; means we start searching at parent_group + 2
because the loop increments group before using it.

for(i = 2 ; i < ngroups ; i++) means we don't search through all the
groups.

The end result is that parent_group and parent_group + 1 are not checked
for free inodes in the final linear search.  ext3 has the same problem.


---

 25-akpm/fs/ext2/ialloc.c |    4 ++--
 25-akpm/fs/ext3/ialloc.c |    4 ++--
 2 files changed, 4 insertions(+), 4 deletions(-)

diff -puN fs/ext2/ialloc.c~ext2-ext3-ENOSPC-fix fs/ext2/ialloc.c
--- 25/fs/ext2/ialloc.c~ext2-ext3-ENOSPC-fix	Tue Mar  2 13:22:56 2004
+++ 25-akpm/fs/ext2/ialloc.c	Tue Mar  2 13:22:56 2004
@@ -431,8 +431,8 @@ static int find_group_other(struct super
 	 * That failed: try linear search for a free inode, even if that group
 	 * has no free blocks.
 	 */
-	group = parent_group + 1;
-	for (i = 2; i < ngroups; i++) {
+	group = parent_group;
+	for (i = 0; i < ngroups; i++) {
 		if (++group >= ngroups)
 			group = 0;
 		desc = ext2_get_group_desc (sb, group, &bh);
diff -puN fs/ext3/ialloc.c~ext2-ext3-ENOSPC-fix fs/ext3/ialloc.c
--- 25/fs/ext3/ialloc.c~ext2-ext3-ENOSPC-fix	Tue Mar  2 13:22:56 2004
+++ 25-akpm/fs/ext3/ialloc.c	Tue Mar  2 13:22:56 2004
@@ -398,8 +398,8 @@ static int find_group_other(struct super
 	 * That failed: try linear search for a free inode, even if that group
 	 * has no free blocks.
 	 */
-	group = parent_group + 1;
-	for (i = 2; i < ngroups; i++) {
+	group = parent_group;
+	for (i = 0; i < ngroups; i++) {
 		if (++group >= ngroups)
 			group = 0;
 		desc = ext3_get_group_desc (sb, group, &bh);

_
